from datetime import time
from enum import Enum
from typing import Optional

from pysungrow.definitions.devices.hybrid import (
    all_hybrid,
    sh3k6,
    sh3k6_30,
    sh4k6,
    sh4k6_30,
    sh5_0rt,
    sh5k_20,
    sh5k_30,
    sh5k_v13,
    sh6_0rt,
    sh8_0rt,
    sh10rt,
)
from pysungrow.definitions.variable import RawType, VariableDefinition, VariableType
from pysungrow.definitions.variables.device import OutputType
from pysungrow.definitions.variables.general import StartStop
from pysungrow.lib.flag import Flag


class SystemState(Enum):
    STOP = 0x0002
    STANDBY = 0x0008
    INITIAL_STANDBY = 0x0010
    STARTUP = 0x0020
    RUNNING = 0x0040
    FAULT = 0x0100
    RUNNING_MAINTAIN_MODE = 0x0400
    RUNNING_FORCED_MODE = 0x0800
    RUNNING_OFFGRID_MODE = 0x1000
    RESTARTING = 0x2501
    RUNNING_EXTERNAL_EMS_MODE = 0x4000


class RunningState(Flag):
    pv_power: bool
    battery_charging: bool
    battery_discharging: bool
    load_is_active: bool
    power_feed_in_the_grid: bool
    importing_power_from_grid: bool
    _reserved: None
    power_generated_from_load: bool


system_state_variable_13000 = VariableDefinition(
    VariableType.READ,
    "system_state",
    13000,
    RawType.U16,
    SystemState,
    devices=all_hybrid,
)
running_state_variable_13001 = VariableDefinition(
    VariableType.READ,
    "running_state",
    13001,
    RawType.U16,
    RunningState,
    devices=all_hybrid,
)
daily_pv_generation_variable_13002 = VariableDefinition(
    VariableType.READ,
    "daily_pv_generation",
    13002,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_pv_generation_variable_13003 = VariableDefinition(
    VariableType.READ,
    "total_pv_generation",
    13003,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
daily_export_from_pv_variable_13005 = VariableDefinition(
    VariableType.READ,
    "daily_export_from_pv",
    13005,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_export_from_pv_variable_13006 = VariableDefinition(
    VariableType.READ,
    "total_export_from_pv",
    13006,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
load_power_variable_13008 = VariableDefinition(
    VariableType.READ,
    "load_power",
    13008,
    RawType.S32,
    int,
    unit="W",
    devices=all_hybrid,
)
export_power_variable_13010 = VariableDefinition(
    VariableType.READ,
    "export_power",
    13010,
    RawType.S32,
    int,
    unit="W",
    devices=all_hybrid,
)
daily_battery_charge_energy_from_pv_variable_13012 = VariableDefinition(
    VariableType.READ,
    "daily_battery_charge_energy_from_pv",
    13012,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_battery_charge_energy_from_pv_variable_13013 = VariableDefinition(
    VariableType.READ,
    "total_battery_charge_energy_from_pv",
    13013,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
co2_reduction_variable_13015 = VariableDefinition(
    VariableType.READ,
    "co2_reduction",
    13015,
    RawType.U32,
    float,
    transform=0.1,
    unit="kg",
    devices=all_hybrid,
)
daily_direct_energy_consumption_variable_13017 = VariableDefinition(
    VariableType.READ,
    "daily_direct_energy_consumption",
    13017,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_direct_energy_consumption_variable_13018 = VariableDefinition(
    VariableType.READ,
    "total_direct_energy_consumption",
    13018,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
battery_voltage_variable_13020 = VariableDefinition(
    VariableType.READ,
    "battery_voltage",
    13020,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices=all_hybrid,
)
battery_current_variable_13021 = VariableDefinition(
    VariableType.READ,
    "battery_current",
    13021,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices=all_hybrid,
)
battery_power_variable_13022 = VariableDefinition(
    VariableType.READ,
    "battery_power",
    13022,
    RawType.U16,
    int,
    unit="W",
    devices=all_hybrid,
)
battery_level_variable_13023 = VariableDefinition(
    VariableType.READ,
    "battery_level",
    13023,
    RawType.U16,
    float,
    transform=0.1,
    unit="%",
    devices=all_hybrid,
)
battery_health_variable_13024 = VariableDefinition(
    VariableType.READ,
    "battery_health",
    13024,
    RawType.U16,
    float,
    transform=0.1,
    unit="%",
    devices=all_hybrid,
)
battery_temperature_variable_13025 = VariableDefinition(
    VariableType.READ,
    "battery_temperature",
    13025,
    RawType.S16,
    float,
    transform=0.1,
    unit="°C",
    devices=all_hybrid,
)
daily_battery_discharge_energy_variable_13026 = VariableDefinition(
    VariableType.READ,
    "daily_battery_discharge_energy",
    13026,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_battery_discharge_energy_variable_13027 = VariableDefinition(
    VariableType.READ,
    "total_battery_discharge_energy",
    13027,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
daily_self_consumption_variable_13029 = VariableDefinition(
    VariableType.READ,
    "daily_self_consumption",
    13029,
    RawType.U16,
    float,
    transform=0.1,
    unit="%",
    devices=all_hybrid,
)


class GridState(Enum):
    OFF_GRID = 0xAA
    ON_GRID = 0x55


grid_state_variable_13030 = VariableDefinition(
    VariableType.READ,
    "grid_state",
    13030,
    RawType.U16,
    GridState,
    transform=lambda raw: None if raw == 0xFFFF else GridState(raw),
    devices=all_hybrid,
)

phase_a_current_variable_13031 = VariableDefinition(
    VariableType.READ,
    "phase_a_current",
    13031,
    RawType.S16,
    float,
    transform=0.1,
    unit="A",
    devices=all_hybrid,
)
phase_b_current_variable_13032 = VariableDefinition(
    VariableType.READ,
    "phase_b_current",
    13032,
    RawType.S16,
    float,
    transform=0.1,
    unit="A",
    devices=all_hybrid,
    if_output_type={OutputType.THREE_PHASE_3P4L, OutputType.THREE_PHASE_3P3L},
)
phase_c_current_variable_13033 = VariableDefinition(
    VariableType.READ,
    "phase_c_current",
    13033,
    RawType.S16,
    float,
    transform=0.1,
    unit="A",
    devices=all_hybrid,
    if_output_type={OutputType.THREE_PHASE_3P4L, OutputType.THREE_PHASE_3P3L},
)

total_active_power_variable_13034 = VariableDefinition(
    VariableType.READ,
    "total_active_power",
    13034,
    RawType.S32,
    int,
    unit="W",
    devices=all_hybrid,
)
daily_import_energy_variable_13036 = VariableDefinition(
    VariableType.READ,
    "daily_import_energy",
    13036,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_import_energy_variable_13037 = VariableDefinition(
    VariableType.READ,
    "total_import_energy",
    13037,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
# note: handled by holding register 13057
# battery_capacity_variable_13039 = VariableDefinition(VariableType.READ, "battery_capacity", 13039, RawType.U16, float, transform=0.1, devices={sh5k_20, sh3k6, sh4k6, sh5k_30, sh5k_v13, sh3k6_30, sh4k6_30})
daily_charge_energy_variable_13040 = VariableDefinition(
    VariableType.READ,
    "daily_charge_energy",
    13040,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_charge_energy_variable_13041 = VariableDefinition(
    VariableType.READ,
    "total_charge_energy",
    13041,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
drm_state_variable_13043 = VariableDefinition(
    VariableType.READ, "drm_state", 13043, RawType.U16, int, devices=all_hybrid
)
daily_export_energy_variable_13045 = VariableDefinition(
    VariableType.READ,
    "daily_export_energy",
    13045,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)
total_export_energy_variable_13046 = VariableDefinition(
    VariableType.READ,
    "total_export_energy",
    13046,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices=all_hybrid,
)

# TODO: alarms and faults (registers 13050-13079 for hybrid)
# TODO: BMS info (registers 13100-13118 for hybrid)

start_stop_variable_13000 = VariableDefinition(
    VariableType.READWRITE,
    "start_stop",
    13000,
    RawType.U16,
    StartStop,
    devices=all_hybrid,
)
battery_maintenance_variable_13001 = VariableDefinition(
    VariableType.READWRITE,
    "battery_maintenance",
    13001,
    RawType.U16,
    bool,
    transform=(lambda raw: raw == 0xAA, lambda val: 0xAA if val else 0x55),
    devices={sh5k_20, sh3k6, sh4k6, sh5k_30, sh5k_v13, sh3k6_30, sh4k6_30},
)


class LoadAdjustmentMode(Enum):
    TIMING = 0
    ON_OFF = 1
    POWER_OPTIMIZED = 2


load_adjustment_mode_variable_13002 = VariableDefinition(
    VariableType.READWRITE,
    "load_adjustment_mode",
    13002,
    RawType.U16,
    LoadAdjustmentMode,
    transform=(
        lambda raw: None
        if not isinstance(raw, int) or raw > 2
        else LoadAdjustmentMode(raw),
        lambda val: val.value if val is not None else 3,
    ),
    devices=all_hybrid,
)
load_period_1_start_variable_13003 = VariableDefinition(
    VariableType.READWRITE,
    "load_period_1_start",
    13003,
    RawType.RAW,
    time,
    length=2,
    transform=(
        lambda raw: time(raw[0], raw[1]) if isinstance(raw, tuple) else None,
        lambda val: (val.hour, val.minute),
    ),
    devices=all_hybrid,
)
load_period_1_end_variable_13005 = VariableDefinition(
    VariableType.READWRITE,
    "load_period_1_end",
    13005,
    RawType.RAW,
    time,
    length=2,
    transform=(
        lambda raw: time(raw[0], raw[1]) if isinstance(raw, tuple) else None,
        lambda val: (val.hour, val.minute),
    ),
    devices=all_hybrid,
)
load_period_2_start_variable_13007 = VariableDefinition(
    VariableType.READWRITE,
    "load_period_2_start",
    13007,
    RawType.RAW,
    time,
    length=2,
    transform=(
        lambda raw: time(raw[0], raw[1]) if isinstance(raw, tuple) else None,
        lambda val: (val.hour, val.minute),
    ),
    devices=all_hybrid,
)
load_period_2_end_variable_13009 = VariableDefinition(
    VariableType.READWRITE,
    "load_period_2_end",
    13009,
    RawType.RAW,
    time,
    length=2,
    transform=(
        lambda raw: time(raw[0], raw[1]) if isinstance(raw, tuple) else None,
        lambda val: (val.hour, val.minute),
    ),
    devices=all_hybrid,
)


class OnOffMode(Enum):
    ON = 0xAA
    OFF = 0x55


load_on_off_mode_variable_13011 = VariableDefinition(
    VariableType.READWRITE,
    "load_on_off_mode",
    13011,
    RawType.U16,
    OnOffMode,
    devices=all_hybrid,
)
load_optimized_start_variable_13012 = VariableDefinition(
    VariableType.READWRITE,
    "load_optimized_start",
    13012,
    RawType.RAW,
    time,
    length=2,
    transform=(
        lambda raw: time(raw[0], raw[1]) if isinstance(raw, tuple) else None,
        lambda val: (val.hour, val.minute),
    ),
    devices=all_hybrid,
)
load_optimized_end_variable_13014 = VariableDefinition(
    VariableType.READWRITE,
    "load_optimized_end",
    13014,
    RawType.RAW,
    time,
    length=2,
    transform=(
        lambda raw: time(raw[0], raw[1]) if isinstance(raw, tuple) else None,
        lambda val: (val.hour, val.minute),
    ),
    devices=all_hybrid,
)
load_optimized_power_variable_13016 = VariableDefinition(
    VariableType.READWRITE,
    "load_optimized_power",
    13016,
    RawType.U16,
    int,
    unit="W",
    limits=(0, 5000),
    devices=all_hybrid,
)


class EMSMode(Enum):
    SELF_CONSUMPTION = 0
    FORCED = 1
    EXTERNAL_EMS = 2


ems_mode_variable_13050 = VariableDefinition(
    VariableType.READWRITE, "ems_mode", 13050, RawType.U16, EMSMode, devices=all_hybrid
)


class ChargeDischargeCommand(Enum):
    CHARGE = 0xAA
    DISCHARGE = 0xBB
    STOP = 0xCC


charge_discharge_command_variable_13051 = VariableDefinition(
    VariableType.READWRITE,
    "charge_discharge_command",
    13051,
    RawType.U16,
    ChargeDischargeCommand,
    devices=all_hybrid,
)
charge_discharge_power_variable_13052 = VariableDefinition(
    VariableType.READWRITE,
    "charge_discharge",
    13052,
    RawType.U16,
    int,
    unit="W",
    limits=(0, 5000),
    devices=all_hybrid,
)


class BatteryType(Enum):
    LEAD_ACID_NARADA = 0
    LI_ION_SAMSUNG = 1
    NO_BATTERY = 2
    LEAD_ACID_OTHER = 3
    LI_ION_US2000A = 4
    LI_ION_LG = 5
    LI_ION_US2000B = 6
    LI_ION_GCL = 7
    LI_ION_BLUESUN = 8
    LI_ION_SUNGROW = 9
    LI_ION_BYD = 10

    @property
    def manufacturer(self) -> Optional[str]:
        """Manufacturer of the battery type, if known."""

        if self == BatteryType.LEAD_ACID_NARADA:
            return "Narada"
        elif self == BatteryType.LI_ION_BYD:
            return "BYD"
        elif self == BatteryType.LI_ION_SAMSUNG:
            return "Samsung"
        elif self == BatteryType.LI_ION_US2000A or self == BatteryType.LI_ION_US2000B:
            return "Pylontech"
        elif self == BatteryType.LI_ION_LG:
            return "LG"
        elif self == BatteryType.LI_ION_GCL:
            return "GCL"
        elif self == BatteryType.LI_ION_BLUESUN:
            return "Bluesun"
        elif self == BatteryType.LI_ION_SUNGROW:
            return "Sungrow"
        else:
            return None


battery_type_variable_13055 = VariableDefinition(
    VariableType.READWRITE,
    "battery_type",
    13055,
    RawType.U16,
    BatteryType,
    devices={sh5k_20, sh5k_30, sh5k_v13, sh3k6, sh3k6_30, sh4k6, sh4k6_30},
)
battery_nominal_voltage_variable_13056 = VariableDefinition(
    VariableType.READWRITE,
    "battery_nominal_voltage",
    13056,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    limits=(30, 60),
    devices={sh5k_20, sh5k_30, sh5k_v13, sh3k6, sh3k6_30, sh4k6, sh4k6_30},
)
battery_capacity_variable_13057 = VariableDefinition(
    VariableType.READWRITE,
    "battery_capacity",
    13057,
    RawType.U16,
    int,
    unit="Ah",
    devices={sh5k_20, sh5k_30, sh5k_v13, sh3k6, sh3k6_30, sh4k6, sh4k6_30},
)
max_soc_variable_13058 = VariableDefinition(
    VariableType.READWRITE,
    "max_soc",
    13058,
    RawType.U16,
    float,
    transform=0.1,
    unit="%",
    limits=(70.0, 100.0),
    devices=all_hybrid,
)
min_soc_variable_13059 = VariableDefinition(
    VariableType.READWRITE,
    "min_soc",
    13059,
    RawType.U16,
    float,
    transform=0.1,
    unit="%",
    limits=(0.0, 50.0),
    devices=all_hybrid,
)
battery_over_voltage_threshold_variable_13060 = VariableDefinition(
    VariableType.READWRITE,
    "battery_over_voltage_threshold",
    13060,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    limits=(48.0, 70.0),
    devices={sh5k_20, sh5k_30, sh5k_v13, sh3k6, sh3k6_30, sh4k6, sh4k6_30},
)
battery_under_voltage_threshold_variable_13061 = VariableDefinition(
    VariableType.READWRITE,
    "battery_under_voltage_threshold",
    13061,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    limits=(32.0, 48.0),
    devices={sh5k_20, sh5k_30, sh5k_v13, sh3k6, sh3k6_30, sh4k6, sh4k6_30},
)
battery_over_temperature_threshold_variable_13062 = VariableDefinition(
    VariableType.READWRITE,
    "battery_over_temperature_threshold",
    13062,
    RawType.U16,
    float,
    transform=0.1,
    unit="°C",
    limits=(20.0, 60.0),
    devices={sh5k_20, sh5k_30, sh5k_v13, sh3k6, sh3k6_30, sh4k6, sh4k6_30},
)
battery_under_temperature_threshold_variable_13063 = VariableDefinition(
    VariableType.READWRITE,
    "battery_under_temperature_threshold",
    13063,
    RawType.U16,
    float,
    transform=0.1,
    unit="°C",
    limits=(-30.0, 10.0),
    devices={sh5k_20, sh5k_30, sh5k_v13, sh3k6, sh3k6_30, sh4k6, sh4k6_30},
)
# TODO: battery parameters in registers 13065-13073 (hybrid)
export_power_limitation_variable_13074 = VariableDefinition(
    VariableType.READWRITE,
    "export_power_limitation",
    13074,
    RawType.U16,
    int,
    unit="W",
    devices=all_hybrid,
)
off_grid_enabled_variable_13075 = VariableDefinition(
    VariableType.READWRITE,
    "off_grid_enabled",
    13075,
    RawType.U16,
    bool,
    transform=(lambda raw: raw == 0xAA, lambda val: 0xAA if val else 0x55),
    devices=all_hybrid,
)
external_ems_heartbeat_variable_13080 = VariableDefinition(
    VariableType.READWRITE,
    "external_ems_heartbeat",
    13080,
    RawType.U16,
    int,
    unit="s",
    limits=(0, 20),
    devices=all_hybrid,
)
export_power_limitation_variable_13087 = VariableDefinition(
    VariableType.READWRITE,
    "export_power_limitation_enabled",
    13087,
    RawType.U16,
    bool,
    transform=(lambda raw: raw == 0xAA, lambda val: 0xAA if val else 0x55),
    devices={sh5_0rt, sh6_0rt, sh8_0rt, sh10rt},
)
reserved_soc_for_backup_variable_13100 = VariableDefinition(
    VariableType.READWRITE,
    "reserved_soc_for_backup",
    13100,
    RawType.U16,
    int,
    unit="%",
    limits=(0, 100),
    devices=all_hybrid,
)



"""Provides the Flag class."""

from typing import Dict, List


class Flag:
    """A class to handle bitfield-style ints representing several flags."""

    def __init__(self, value: int):
        """
        Construct an instance of this flag class, including parsing bits into the defined flags.

        :param value: Bitfield-like integer to extract bits from
        """
        # __annotations__ assumed to be ordered: https://stackoverflow.com/a/64446226
        for index, entry in enumerate(self.__annotations__.keys()):
            if entry.startswith("_"):
                continue
            self.__dict__[entry] = bool(value & (1 << index))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(f'{e}={self.__dict__[e]}' for e in self.keys())})"

    def keys(self) -> List[str]:
        """Return the keys of the flags handled by this class."""
        return [k for k in self.__annotations__.keys() if not k.startswith("_")]

    def to_dict(self) -> Dict[str, bool]:
        """Return the flags with their values in this class."""
        return {k: self.__dict__[k] for k in self.keys()}



from typing import List

from pysungrow.definitions.variable import VariableDefinition

from . import device, general, hybrid, string

variables: List[VariableDefinition] = [
    v
    for module in [device, general, hybrid, string]
    for v in module.__dict__.values()
    if isinstance(v, VariableDefinition)
]



"""Main access point for connecting to inverters."""

from typing import Any, Dict, List, Optional, Sequence

from pymodbus.client.base import ModbusBaseClient

from pysungrow.definitions.device import SungrowDevice
from pysungrow.definitions.variable import VariableDefinition, VariableType
from pysungrow.definitions.variables import variables
from pysungrow.definitions.variables.device import OutputType
from pysungrow.lib.read_variables import read_variables
from pysungrow.lib.write_variable import write_variable


class SungrowClient:
    """Provides access to variables of the given device. Caches fetched values."""

    def __init__(
        self,
        client: ModbusBaseClient,
        device: SungrowDevice,
        output_type: OutputType,
        slave: int = 1,
    ):
        """
        Construct a client for communicating with Sungrow inverters.

        :param client: The Modbus client over which communication takes place
        :param device: The Sungrow device model which will be communicated with, controls the available variables
        :param output_type: The output type configured for the inverter, controls some available variables
        :param slave: The unit/slave to connect to
        """
        self._client = client
        self._device = device
        self._output_type = output_type
        self._slave = slave
        self._data: Dict[str, Any] = {}

    @property
    def device(self) -> SungrowDevice:
        """Device type connected to."""
        return self._device

    @property
    def data(self) -> Dict[str, Any]:
        """Access to cached data."""
        return self._data

    @property
    def _variables(self):
        return (
            r
            for r in variables
            if self._device in r.devices
            and (r.if_output_type is None or self._output_type in r.if_output_type)
        )

    @property
    def keys(self) -> List[str]:
        """All keys available for the device."""
        return [r.key for r in self._variables]

    def variable(self, key: str) -> Optional[VariableDefinition]:
        """
        Retrieve the variable definition of a variable given its key.

        :param key: Key of the variable
        :return: The variable definition
        """
        return next((r for r in self._variables if r.key == key), None)

    def _variable_should_refresh(
        self, keys: Optional[Sequence[str]], variable: VariableDefinition
    ):
        if keys is not None and variable.key not in keys:
            return False
        if (
            variable.variable_type == VariableType.READ_STATIC
            and variable.key in self._data
        ):
            return False
        return True

    async def refresh(self, keys: Optional[Sequence[str]] = None):
        """
        Refresh the cached values of variables for the device.

        :param keys: Optionally restrict the variables to refresh to this set
        """
        _, conn = await self._client.connect()
        try:
            await self._refresh_impl(conn, keys)
        finally:
            await conn.close()

    async def _refresh_impl(
        self, conn: ModbusBaseClient, keys: Optional[Sequence[str]]
    ):
        self._data.update(
            await read_variables(
                conn,
                [v for v in self._variables if self._variable_should_refresh(keys, v)],
                self._slave,
            )
        )

    async def get(self, key: str) -> Any:
        """
        Retrieve the (cached) value for a variable with the given key. Fetches data if no cached value is available.

        :param key: The key of the variable to retrieve
        :return: The (cached) value
        """
        if key not in self.keys:
            raise KeyError("The given key is not available in this device")
        if key not in self._data:
            await self.refresh()
        return self._data[key]

    async def set(self, key: str, value: Any):
        """
        Set the value of the variable with the given key.

        :param key: Key of the variable to set
        :param value: New value
        """
        variable = self.variable(key)
        if not variable:
            raise KeyError("The given key is not available in this device")
        elif variable.variable_type != VariableType.READWRITE:
            raise KeyError("Given key is not writeable")
        elif variable.type != type(value):
            raise ValueError(f"Value for given key must be of type {variable.type}")

        _, conn = await self._client.connect()
        try:
            await write_variable(conn, value, variable, self._slave)
            await self._refresh_impl(conn, [key])
        finally:
            await conn.close()



from argparse import ArgumentParser
import asyncio
from enum import Enum

from pymodbus.client import AsyncModbusTcpClient

from pysungrow import SungrowClient, identify
from pysungrow.definitions.device import SungrowDevice


async def do_identify(host, port, slave):
    client = AsyncModbusTcpClient(host, port)
    serial_number, device, output_type = await identify(client, slave)
    print("Serial number:", serial_number)
    print("Device:", device.name)
    print("Output type:", output_type.name)


async def do_get(host, port, slave, keys):
    client = AsyncModbusTcpClient(host, port)
    _, device, output_type = await identify(client, slave)
    sungrow = SungrowClient(client, device, output_type, slave)
    await sungrow.refresh()
    for key in sungrow.keys:
        if keys is not None and key not in keys:
            continue

        value = await sungrow.get(key)
        if isinstance(value, SungrowDevice):
            value = value.name
        unit = sungrow.variable(key).unit
        print(key + ":", f"{value} {unit}" if unit is not None else value)


async def do_set(host, port, slave, key, value):
    client = AsyncModbusTcpClient(host, port)
    _, device, output_type = await identify(client, slave)
    sungrow = SungrowClient(client, device, output_type, slave)

    variable = sungrow.variable(key)
    if variable.type in (int, float):
        value = variable.type(value)
    elif issubclass(variable.type, Enum):
        value = variable.type[value]
    elif variable.type == bool:
        if value.lower() == "true" or value == "1":
            value = True
        elif value.lower() == "false" or value == "0":
            value = False
        else:
            raise ValueError(
                "Value for boolean types must be one of 'true', 'false', '1' or '0'"
            )

    old = await sungrow.get(key)
    unit = variable.unit
    print("Current value:", f"{old} {unit}" if unit is not None else old)
    print("Setting value:", f"{value} {unit}" if unit is not None else value)
    await sungrow.set(key, value)
    new = await sungrow.get(key)
    print("New value:", f"{new} {unit}" if unit is not None else new)


parser = ArgumentParser(prog="pysungrow")
parser.add_argument("host")
parser.add_argument("-p", "--port", type=int, default=502)
parser.add_argument("-s", "--slave", type=int, default=1)

subparsers = parser.add_subparsers()
parser_identify = subparsers.add_parser("identify")
parser_identify.set_defaults(
    func=lambda arg: asyncio.run(do_identify(arg.host, arg.port, arg.slave))
)
parser_get = subparsers.add_parser("get")
parser_get.add_argument("-k", "--key", action="append")
parser_get.set_defaults(
    func=lambda arg: asyncio.run(do_get(arg.host, arg.port, arg.slave, arg.key))
)
parser_set = subparsers.add_parser("set")
parser_set.add_argument("key")
parser_set.add_argument("value")
parser_set.set_defaults(
    func=lambda arg: asyncio.run(
        do_set(arg.host, arg.port, arg.slave, arg.key, arg.value)
    )
)

args = parser.parse_args()
args.func(args)



from enum import Enum
from typing import Optional, Tuple, Union

from pysungrow.definitions.device import SungrowDevice
from pysungrow.definitions.devices.hybrid import all_hybrid
from pysungrow.definitions.devices.string import (
    all_string,
    sg8ktl_m,
    sg10ktl_m,
    sg12ktl_m,
    sg80ktl_m,
    sg125hv,
)
from pysungrow.definitions.variable import RawType, VariableDefinition, VariableType


def _get_sungrow_device(
    code: Union[int, str, Tuple[int, ...]]
) -> Optional[SungrowDevice]:
    assert isinstance(code, int)

    from pysungrow.definitions.devices import devices

    return next((d for d in devices if d.code == code), None)


protocol_number_variable_4950 = VariableDefinition(
    VariableType.READ_STATIC,
    "protocol_number",
    4950,
    RawType.U32,
    int,
    devices={*all_string, *all_hybrid},
)
protocol_version_variable_4952 = VariableDefinition(
    VariableType.READ_STATIC,
    "protocol_version",
    4952,
    RawType.U32,
    int,
    devices={*all_string, *all_hybrid},
)
arm_software_version_variable_4954 = VariableDefinition(
    VariableType.READ_STATIC,
    "arm_software_version",
    4954,
    RawType.STRING,
    str,
    length=15,
    devices={sg8ktl_m, sg10ktl_m, sg12ktl_m, sg80ktl_m, sg125hv, *all_hybrid},
)
dsp_software_version_variable_4969 = VariableDefinition(
    VariableType.READ_STATIC,
    "dsp_software_version",
    4969,
    RawType.STRING,
    str,
    length=15,
    devices={sg8ktl_m, sg10ktl_m, sg12ktl_m, sg80ktl_m, sg125hv, *all_hybrid},
)
serial_number_variable_4990 = VariableDefinition(
    VariableType.READ_STATIC,
    "serial_number",
    4990,
    RawType.STRING,
    str,
    length=10,
    devices={*all_string, *all_hybrid},
)
device_type_variable_5000 = VariableDefinition(
    VariableType.READ_STATIC,
    "device_type",
    5000,
    RawType.U16,
    SungrowDevice,
    transform=_get_sungrow_device,
    devices={*all_string, *all_hybrid},
)
nominal_output_power_variable_5001 = VariableDefinition(
    VariableType.READ_STATIC,
    "nominal_output_power",
    5001,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    devices={*all_string, *all_hybrid},
)


class OutputType(Enum):
    """The output type of an inverter, as configured."""

    SINGLE_PHASE = 0
    THREE_PHASE_3P4L = 1
    THREE_PHASE_3P3L = 2


output_type_variable_5002 = VariableDefinition(
    VariableType.READ_STATIC,
    "output_type",
    5002,
    RawType.U16,
    OutputType,
    transform=OutputType,
    devices={*all_string, *all_hybrid},
)



"""Functions for reading from Modbus clients."""

from typing import Any, Dict, List

from pymodbus.client import ModbusBaseClient

from pysungrow.definitions.variable import VariableDefinition, VariableType
from pysungrow.lib.decode import decode_variables
from pysungrow.lib.group_variables_by_proximity import group_variables_by_proximity


async def read_variables(
    client: ModbusBaseClient, variables: List[VariableDefinition], slave: int
) -> Dict[str, Any]:
    """
    Read some registers from a client.

    Batches register read requests together for better efficiency and decodes the results.

    :param client: The client from which to read
    :param variables: The registers that should be read
    :param slave: The unit/slave from which to read
    :return: The read and decoded registers
    """
    input_groups = group_variables_by_proximity(
        [
            r
            for r in variables
            if r.variable_type in (VariableType.READ, VariableType.READ_STATIC)
        ]
    )
    holding_groups = group_variables_by_proximity(
        [r for r in variables if r.variable_type == VariableType.READWRITE]
    )

    result: Dict[str, Any] = {}
    for group in input_groups:
        # note: communication address = protocol address - 1
        response = await client.read_input_registers(
            group[0].address - 1,
            count=group[-1].end_address - group[0].address + 1,
            slave=slave,
        )
        result.update(decode_variables(response.registers, group))
    for group in holding_groups:
        # note: communication address = protocol address - 1
        response = await client.read_holding_registers(
            group[0].address - 1,
            count=group[-1].end_address - group[0].address + 1,
            slave=slave,
        )
        result.update(decode_variables(response.registers, group))
    return result



from datetime import datetime, timedelta
from enum import Enum

from pysungrow.definitions.devices.string import (
    all_string,
    sg8ktl_m,
    sg10ktl_m,
    sg12ktl_m,
    sg33ktl_m,
    sg36ktl_m,
    sg40ktl_m,
    sg49k5j,
    sg50ktl_m,
    sg60ktl,
    sg60ktl_m,
    sg60ku,
    sg60ku_m,
    sg80hv,
    sg80ktl,
    sg80ktl_m,
    sg125hv,
)
from pysungrow.definitions.variable import RawType, VariableDefinition, VariableType
from pysungrow.definitions.variables.general import StartStop

grid_frequency_variable_5148 = VariableDefinition(
    VariableType.READ,
    "grid_frequency_fine",
    5148,
    RawType.U16,
    float,
    transform=0.01,
    unit="Hz",
    devices={sg8ktl_m, sg10ktl_m, sg12ktl_m, sg80ktl_m, sg80hv, sg125hv},
)


class SystemState(Enum):
    RUNNING = 0x0
    STOP = 0x8000
    STOP_KEY = 0x1300
    STOP_EMERGENCY = 0x1500
    STANDBY = 0x1400
    INITIAL_STANDBY = 0x1200
    STARTUP = 0x1600
    RUNNING_ALARM = 0x9100
    RUNNING_DERATING = 0x8100
    RUNNING_DISPATCH = 0x8200
    RUNNING_FAULT = 0x5500
    COMMUNICATE_FAULT = 0x2500


system_state_variable_5038 = VariableDefinition(
    VariableType.READ,
    "system_state",
    5038,
    RawType.U16,
    SystemState,
    devices=all_string,
)
fault_date_variable_5039 = VariableDefinition(
    VariableType.READ,
    "fault_date",
    5039,
    RawType.RAW,
    datetime,
    transform=lambda raw: datetime(
        year=raw[0], month=raw[1], day=raw[2], hour=raw[3], minute=raw[4], second=raw[5]
    )
    if isinstance(raw, tuple) and raw[0] > 0
    else None,
    length=6,
    devices={*all_string},
)
fault_code_variable_5045 = VariableDefinition(
    VariableType.READ, "fault_code", 5045, RawType.U16, int, devices={*all_string}
)
nominal_reactive_output_power_variable_5049 = VariableDefinition(
    VariableType.READ,
    "nominal_reactive_output_power",
    5049,
    RawType.U16,
    float,
    transform=0.1,
    unit="K var",
    devices={*all_string},
)
impedance_to_ground_variable_5071 = VariableDefinition(
    VariableType.READ,
    "impedance_to_ground",
    5071,
    RawType.U16,
    int,
    unit="kΩ",
    devices={*all_string},
)
work_state_variable_5081 = VariableDefinition(
    VariableType.READ, "work_state2", 5081, RawType.U32, int, devices={*all_string}
)
daily_running_time_variable_5113 = VariableDefinition(
    VariableType.READ,
    "daily_running_time",
    5113,
    RawType.U16,
    timedelta,
    transform=lambda minutes: timedelta(minutes=minutes)
    if isinstance(minutes, int)
    else None,
    devices={*all_string},
)


class CountryCode(Enum):
    GB = 0
    DE = 1
    FR = 2
    IT = 3
    ES = 4
    AT = 5
    AU = 6
    CZ = 7
    BE = 8
    DK = 9
    GR_L = 10
    GR_IS = 11
    NL = 12
    PT = 13
    CHN = 14
    SE = 15
    OTHER_50HZ = 16
    RO = 17
    TH = 18
    TK = 19
    AU_WEST = 20
    VORARLBERG_AUSTRIA = 25
    CA = 60
    US = 61
    OTHER_60HZ = 62
    JP_50HZ = 70
    JP_60HZ = 71


country_variable_5114 = VariableDefinition(
    VariableType.READ_STATIC,
    "country",
    5114,
    RawType.U16,
    CountryCode,
    transform=CountryCode,
    devices={*all_string},
)
monthly_power_yield_variable_5128 = VariableDefinition(
    VariableType.READ,
    "monthly_power_yield",
    5128,
    RawType.U32,
    float,
    transform=0.1,
    unit="kWh",
    devices={*all_string},
)
negative_voltage_to_ground_5146 = VariableDefinition(
    VariableType.READ,
    "negative_voltage_to_ground",
    5146,
    RawType.S16,
    float,
    transform=0.1,
    unit="V",
    devices={*all_string},
)
bus_voltage_variable_5147 = VariableDefinition(
    VariableType.READ,
    "bus_voltage",
    5147,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices={*all_string},
)

start_stop_variable_5006 = VariableDefinition(
    VariableType.READWRITE,
    "start_stop",
    5006,
    RawType.U16,
    StartStop,
    devices=all_string,
)
power_limitation_variable_5007 = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation",
    5007,
    RawType.U16,
    bool,
    transform=(lambda raw: raw == 0xAA, lambda val: 0xAA if val else 0x55),
    devices=all_string,
)
power_limitation_setting_variable_5008_a = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_setting",
    5008,
    RawType.U16,
    float,
    transform=0.1,
    unit="%",
    limits=(0, 110),
    devices={
        sg50ktl_m,
        sg60ktl_m,
        sg60ktl,
        sg60ku_m,
        sg60ku,
        sg33ktl_m,
        sg40ktl_m,
        sg80ktl_m,
    },
)
power_limitation_setting_variable_5008_b = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_setting",
    5008,
    RawType.U16,
    float,
    transform=0.1,
    unit="%",
    limits=(0, 100),
    devices={sg49k5j, sg36ktl_m, sg10ktl_m, sg12ktl_m, sg80ktl, sg125hv},
)
power_factor_setting_variable_5019 = VariableDefinition(
    VariableType.READWRITE,
    "power_factor_setting",
    5019,
    RawType.S16,
    float,
    transform=0.001,
    limits=(-1, 1),
    devices=all_string,
)


class ReactivePowerAdjustment(Enum):
    OFF = 0x55
    SET_POWER_FACTOR = 0xA1
    SET_REACTIVE_PERCENTAGE = 0xA2
    ENABLE_QP_CURVE = 0xA3
    ENABLE_QU_CURVE = 0xA4


reactive_power_adjustment_switch_variable_5036 = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment_switch",
    5036,
    RawType.U16,
    ReactivePowerAdjustment,
    devices=all_string,
)
reactive_power_percentage_variable_5037 = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_percentage",
    5037,
    RawType.S16,
    float,
    transform=0.1,
    unit="%",
    limits=(-100.0, 100.0),
    devices=all_string,
)
power_limitation_adjustment_variable_5039_a = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 55.0),
    devices={sg50ktl_m},
)
power_limitation_adjustment_variable_5039_b = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 66.0),
    devices={sg60ktl_m, sg60ku_m, sg60ktl, sg60ku},
)
power_limitation_adjustment_variable_5039_c = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 49.5),
    devices={sg49k5j},
)
power_limitation_adjustment_variable_5039_d = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 36.3),
    devices={sg33ktl_m},
)
power_limitation_adjustment_variable_5039_e = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 44.0),
    devices={sg40ktl_m},
)
power_limitation_adjustment_variable_5039_f = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 88.0),
    devices={sg80ktl_m},
)
power_limitation_adjustment_variable_5039_g = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 36.0),
    devices={sg36ktl_m},
)
power_limitation_adjustment_variable_5039_h = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 10.0),
    devices={sg10ktl_m},
)
power_limitation_adjustment_variable_5039_i = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 12.0),
    devices={sg125hv},
)
power_limitation_adjustment_variable_5039_j = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 80.0),
    devices={sg80ktl},
)
power_limitation_adjustment_variable_5039_k = VariableDefinition(
    VariableType.READWRITE,
    "power_limitation_adjustment",
    5039,
    RawType.U16,
    float,
    transform=0.1,
    unit="kW",
    limits=(0.0, 125.0),
    devices={sg125hv},
)
reactive_power_adjustment_variable_5040_a = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-25.0, 25.0),
    devices={sg50ktl_m},
)
reactive_power_adjustment_variable_5040_b = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-30.0, 30.0),
    devices={sg60ktl_m, sg60ku_m, sg60ktl, sg60ku},
)
reactive_power_adjustment_variable_5040_c = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-24.7, 24.7),
    devices={sg49k5j},
)
reactive_power_adjustment_variable_5040_d = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-16.5, 16.5),
    devices={sg33ktl_m},
)
reactive_power_adjustment_variable_5040_e = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-20.0, 20.0),
    devices={sg40ktl_m},
)
reactive_power_adjustment_variable_5040_f = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-18.0, 18.0),
    devices={sg36ktl_m},
)
reactive_power_adjustment_variable_5040_g = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-5.0, 5.0),
    devices={sg10ktl_m},
)
reactive_power_adjustment_variable_5040_h = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-6.0, 6.0),
    devices={sg12ktl_m},
)
reactive_power_adjustment_variable_5040_i = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-40.0, 40.0),
    devices={sg80ktl, sg80ktl_m},
)
reactive_power_adjustment_variable_5040_j = VariableDefinition(
    VariableType.READWRITE,
    "reactive_power_adjustment",
    5040,
    RawType.S16,
    float,
    transform=0.1,
    unit="K var",
    limits=(-62.0, 62.0),
    devices={sg125hv},
)



from pysungrow.definitions.device import SungrowDevice, SungrowDeviceType

sh5k_20 = SungrowDevice(0xD09, "SH5K-20", SungrowDeviceType.HYBRID)
sh3k6 = SungrowDevice(
    0xD06,
    "SH3K6",
    SungrowDeviceType.HYBRID,
)
sh4k6 = SungrowDevice(
    0xD07,
    "SH4K6",
    SungrowDeviceType.HYBRID,
)
sh5k_v13 = SungrowDevice(
    0xD03,
    "SH5K-V13",
    SungrowDeviceType.HYBRID,
)
sh5k_30 = SungrowDevice(
    0xD0C,
    "SH5K-30",
    SungrowDeviceType.HYBRID,
)
sh3k6_30 = SungrowDevice(
    0xD0A,
    "SH3K6-30",
    SungrowDeviceType.HYBRID,
)
sh4k6_30 = SungrowDevice(
    0xD0B,
    "SH4K6-30",
    SungrowDeviceType.HYBRID,
)
sh5_0rs = SungrowDevice(
    0xD0F,
    "SH5.0RS",
    SungrowDeviceType.HYBRID,
)
sh3_6rs = SungrowDevice(
    0xD0D,
    "SH3.6RS",
    SungrowDeviceType.HYBRID,
)
sh4_6rs = SungrowDevice(
    0xD0E,
    "SH4.6RS",
    SungrowDeviceType.HYBRID,
)
sh6_0rs = SungrowDevice(
    0xD10,
    "SH6.0RS",
    SungrowDeviceType.HYBRID,
)
sh10rt = SungrowDevice(
    0xE03,
    "SH10RT",
    SungrowDeviceType.HYBRID,
)
sh8_0rt = SungrowDevice(
    0xE02,
    "SH8.0RT",
    SungrowDeviceType.HYBRID,
)
sh6_0rt = SungrowDevice(
    0xE01,
    "SH6.0RT",
    SungrowDeviceType.HYBRID,
)
sh5_0rt = SungrowDevice(
    0xE00,
    "SH5.0RT",
    SungrowDeviceType.HYBRID,
)

all_hybrid = {
    sh5k_20,
    sh3k6,
    sh4k6,
    sh5k_v13,
    sh5k_30,
    sh3k6_30,
    sh4k6_30,
    sh5_0rs,
    sh3_6rs,
    sh4_6rs,
    sh6_0rs,
    sh10rt,
    sh8_0rt,
    sh6_0rt,
    sh5_0rt,
}



from datetime import datetime
from enum import Enum

from pysungrow.definitions.devices.hybrid import all_hybrid
from pysungrow.definitions.devices.string import (
    all_string,
    sg33ktl_m,
    sg36ktl_m,
    sg40ktl_m,
    sg49k5j,
    sg50ktl_m,
    sg60ktl_m,
    sg60ku_m,
    sg80ktl_m,
)
from pysungrow.definitions.variable import RawType, VariableDefinition, VariableType
from pysungrow.definitions.variables.device import OutputType

daily_output_energy_variable_5003 = VariableDefinition(
    VariableType.READ,
    "daily_output_energy",
    5003,
    RawType.U16,
    float,
    transform=0.1,
    unit="kWh",
    devices={*all_string, *all_hybrid},
)
total_output_energy_variable_5004 = VariableDefinition(
    VariableType.READ,
    "total_output_energy",
    5004,
    RawType.U32,
    int,
    unit="kWh",
    devices={*all_string, *all_hybrid},
)
total_running_time_variable_5006 = VariableDefinition(
    VariableType.READ,
    "total_running_time",
    5006,
    RawType.U32,
    int,
    unit="h",
    devices={*all_string},
)
internal_temperature_variable_5008 = VariableDefinition(
    VariableType.READ,
    "internal_temperature",
    5008,
    RawType.S16,
    float,
    transform=0.1,
    unit="°C",
    devices={*all_string, *all_hybrid},
)


mppt_1_voltage_variable_5011 = VariableDefinition(
    VariableType.READ,
    "mppt_1_voltage",
    5011,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices={*all_hybrid},
)
mppt_1_current_variable_5012 = VariableDefinition(
    VariableType.READ,
    "mppt_1_current",
    5012,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices={*all_hybrid},
)
mppt_2_voltage_variable_5013 = VariableDefinition(
    VariableType.READ,
    "mppt_2_voltage",
    5013,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices={*all_hybrid},
)
mppt_2_current_variable_5014 = VariableDefinition(
    VariableType.READ,
    "mppt_2_current",
    5014,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices={*all_hybrid},
)
dc_1_voltage_variable_5011 = VariableDefinition(
    VariableType.READ,
    "dc_1_voltage",
    5011,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices={*all_string},
)
dc_1_current_variable_5012 = VariableDefinition(
    VariableType.READ,
    "dc_1_current",
    5012,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices={*all_string},
)
dc_2_voltage_variable_5013 = VariableDefinition(
    VariableType.READ,
    "dc_2_voltage",
    5013,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices={
        sg33ktl_m,
        sg40ktl_m,
        sg50ktl_m,
        sg60ktl_m,
        sg60ku_m,
        sg49k5j,
        sg80ktl_m,
        sg36ktl_m,
    },
)
dc_2_current_variable_5014 = VariableDefinition(
    VariableType.READ,
    "dc_2_current",
    5014,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices={
        sg33ktl_m,
        sg40ktl_m,
        sg50ktl_m,
        sg60ktl_m,
        sg60ku_m,
        sg49k5j,
        sg80ktl_m,
        sg36ktl_m,
    },
)
dc_3_voltage_variable_5015 = VariableDefinition(
    VariableType.READ,
    "dc_3_voltage",
    5015,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices={
        sg33ktl_m,
        sg40ktl_m,
        sg50ktl_m,
        sg60ktl_m,
        sg60ku_m,
        sg49k5j,
        sg80ktl_m,
        sg36ktl_m,
    },
)
dc_3_current_variable_5016 = VariableDefinition(
    VariableType.READ,
    "dc_3_current",
    5016,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices={
        sg33ktl_m,
        sg40ktl_m,
        sg50ktl_m,
        sg60ktl_m,
        sg60ku_m,
        sg49k5j,
        sg80ktl_m,
        sg36ktl_m,
    },
)
dc_4_voltage_variable_5115 = VariableDefinition(
    VariableType.READ,
    "dc_4_voltage",
    5115,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    devices={sg49k5j, sg50ktl_m, sg60ktl_m, sg60ku_m, sg80ktl_m},
)
dc_4_current_variable_5116 = VariableDefinition(
    VariableType.READ,
    "dc_4_current",
    5116,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices={sg49k5j, sg50ktl_m, sg60ktl_m, sg60ku_m, sg80ktl_m},
)
total_dc_power_variable_5017 = VariableDefinition(
    VariableType.READ,
    "total_dc_power",
    5017,
    RawType.U32,
    int,
    unit="W",
    devices={*all_string, *all_hybrid},
)
phase_a_voltage_variable_5019 = VariableDefinition(
    VariableType.READ,
    "phase_a_voltage",
    5019,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    if_output_type={OutputType.SINGLE_PHASE, OutputType.THREE_PHASE_3P4L},
    devices={*all_string, *all_hybrid},
)
phase_b_voltage_variable_5020 = VariableDefinition(
    VariableType.READ,
    "phase_b_voltage",
    5020,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    if_output_type={OutputType.THREE_PHASE_3P4L},
    devices={*all_string, *all_hybrid},
)
phase_c_voltage_variable_5021 = VariableDefinition(
    VariableType.READ,
    "phase_c_voltage",
    5021,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    if_output_type={OutputType.THREE_PHASE_3P4L},
    devices={*all_string, *all_hybrid},
)
line_ab_voltage_variable_5019 = VariableDefinition(
    VariableType.READ,
    "line_ab_voltage",
    5019,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    if_output_type={OutputType.THREE_PHASE_3P3L},
    description="A-B line voltage",
    devices={*all_string, *all_hybrid},
)
line_bc_voltage_variable_5020 = VariableDefinition(
    VariableType.READ,
    "line_bc_voltage",
    5020,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    if_output_type={OutputType.THREE_PHASE_3P3L},
    description="B-C line voltage",
    devices={*all_string, *all_hybrid},
)
line_ca_voltage_variable_5021 = VariableDefinition(
    VariableType.READ,
    "line_ca_voltage",
    5021,
    RawType.U16,
    float,
    transform=0.1,
    unit="V",
    if_output_type={OutputType.THREE_PHASE_3P3L},
    description="C-A line voltage",
    devices={*all_string, *all_hybrid},
)
phase_a_current_variable_5022 = VariableDefinition(
    VariableType.READ,
    "phase_a_current",
    5022,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    devices={*all_string},
)
phase_b_current_variable_5023 = VariableDefinition(
    VariableType.READ,
    "phase_b_current",
    5023,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    if_output_type={OutputType.THREE_PHASE_3P4L, OutputType.THREE_PHASE_3P3L},
    devices={*all_string},
)
phase_c_current_variable_5024 = VariableDefinition(
    VariableType.READ,
    "phase_c_current",
    5024,
    RawType.U16,
    float,
    transform=0.1,
    unit="A",
    if_output_type={OutputType.THREE_PHASE_3P4L, OutputType.THREE_PHASE_3P3L},
    devices={*all_string},
)
total_active_power_variable_5031 = VariableDefinition(
    VariableType.READ,
    "total_active_power",
    5031,
    RawType.U32,
    int,
    unit="W",
    devices={*all_string},
)
reactive_power_variable_5033 = VariableDefinition(
    VariableType.READ,
    "reactive_power",
    5033,
    RawType.S32,
    int,
    unit="var",
    devices={*all_string, *all_hybrid},
)
power_factor_variable_5035 = VariableDefinition(
    VariableType.READ,
    "power_factor",
    5035,
    RawType.S16,
    float,
    transform=0.001,
    devices={*all_string, *all_hybrid},
)
grid_frequency_variable_5036 = VariableDefinition(
    VariableType.READ,
    "grid_frequency",
    5036,
    RawType.U16,
    float,
    transform=0.1,
    unit="Hz",
    devices={*all_string, *all_hybrid},
)

system_clock_variable_5000 = VariableDefinition(
    VariableType.READWRITE,
    "system_clock",
    5000,
    RawType.RAW,
    datetime,
    transform=(
        lambda raw: datetime(
            year=raw[0],
            month=raw[1],
            day=raw[2],
            hour=raw[3],
            minute=raw[4],
            second=raw[5],
        )
        if isinstance(raw, tuple)
        else None,
        lambda dt: (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second),
    ),
    length=6,
    devices={*all_string, *all_hybrid},
)


class StartStop(Enum):
    START = 0xCF
    STOP = 0xCE
    EMERGENCY_STOP = 0xBB



"""Function to group variables for batching."""

from typing import List

from pysungrow.definitions.variable import VariableDefinition


def group_variables_by_proximity(
    variables: List[VariableDefinition], max_distance: int = 100
) -> List[List[VariableDefinition]]:
    """
    Group variables together if they are within a given number of addresses between each other.

    Used to batch register read calls together.

    :param variables: The registers that should be grouped
    :param max_distance: The maximum distance between registers to be put into the same group.
    :return: List of groups of registers
    """
    groups: List[List[VariableDefinition]] = [[]]
    variables = sorted(variables, key=lambda r: r.address)
    for variable in variables:
        if (
            len(groups[-1]) > 0
            and (variable.address - groups[-1][-1].end_address) > max_distance
        ):
            groups.append([])
        groups[-1].append(variable)
    if len(groups[-1]) == 0:
        groups.pop()
    return groups



"""Classes for defining a Sungrow inverter."""

from dataclasses import dataclass
from enum import Enum


class SungrowDeviceType(str, Enum):
    """Type of inverter."""

    HYBRID = "Hybrid"
    STRING = "String"


@dataclass
class SungrowDevice:
    """Basic definition of a Sungrow inverter."""

    code: int
    name: str
    device_type: SungrowDeviceType

    def __hash__(self):
        return self.code



"""Functions for encoding data to registers."""

from enum import Enum
from typing import List, Tuple, TypeVar, Union

from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadBuilder

from pysungrow.definitions.variable import RawType, VariableDefinition

T = TypeVar("T")


def encode_registers_for_variable(
    data: T, variable: VariableDefinition[T]
) -> List[int]:
    """
    Encode the Python-native data of a variable to Modbus registers.

    :param data: The Python-native data to encode
    :param variable: Description of how the data should be encoded
    :return: The resulting registers
    """
    if variable.type != type(data):
        raise TypeError()

    builder = BinaryPayloadBuilder(wordorder=Endian.Little, byteorder=Endian.Big)

    raw: Union[int, str, Tuple[int, ...]]
    if isinstance(variable.transform, tuple):
        raw = variable.transform[1](data)
    elif isinstance(variable.transform, (float, int)) and variable.type in (float, int):
        assert isinstance(data, (float, int))
        raw = int(round(data / variable.transform))
    elif variable.type in (float, int):
        assert isinstance(data, (float, int))
        raw = int(round(data))
    elif issubclass(variable.type, Enum):
        assert variable.type == type(data)
        raw = data.value
    elif isinstance(data, int) and variable.raw_type in (
        RawType.U16,
        RawType.U32,
        RawType.S16,
        RawType.S32,
    ):
        raw = data
    else:  # pragma: no cover
        raise ValueError("Cannot encode data")

    if variable.raw_type == RawType.U16:
        builder.add_16bit_uint(raw)
    elif variable.raw_type == RawType.U32:
        builder.add_32bit_uint(raw)
    elif variable.raw_type == RawType.S16:
        builder.add_16bit_int(raw)
    elif variable.raw_type == RawType.S32:
        builder.add_32bit_int(raw)
    elif variable.raw_type == RawType.STRING:
        raise NotImplementedError()
    elif variable.raw_type == RawType.RAW:
        assert variable.length is not None
        if not isinstance(raw, tuple):
            raise TypeError()
        for word in raw:
            builder.add_16bit_uint(word)
    else:  # pragma: no cover
        raise KeyError("Unknown raw type")

    return builder.to_registers()



from typing import List

from pysungrow.definitions.device import SungrowDevice

from . import hybrid, string

devices: List[SungrowDevice] = [
    v
    for module in [hybrid, string]
    for v in module.__dict__.values()
    if isinstance(v, SungrowDevice)
]



"""Identifying the type/model of inverter."""

from typing import Tuple

from pymodbus.client import ModbusBaseClient
from pymodbus.exceptions import ModbusException

from pysungrow.definitions.device import SungrowDevice
from pysungrow.definitions.variables.device import (
    OutputType,
    device_type_variable_5000,
    output_type_variable_5002,
    serial_number_variable_4990,
)
from pysungrow.lib.read_variables import read_variables


class NotASungrowDeviceException(BaseException):
    """Thrown when attempting to communicate with a device that is not recognized as a Sungrow inverter."""


async def identify(
    client: ModbusBaseClient, slave: int = 1
) -> Tuple[str, SungrowDevice, OutputType]:
    """
    Identify the Sungrow device to which the client is connected.

    :param client: Modbus client with which to connect
    :param slave: Unit/slave identifier
    :return: Tuple of serial number, device and output type
    """
    _, conn = await client.connect()
    try:
        result = await read_variables(
            conn,
            [
                serial_number_variable_4990,
                device_type_variable_5000,
                output_type_variable_5002,
            ],
            slave,
        )
        return result["serial_number"], result["device_type"], result["output_type"]
    except ModbusException:
        raise NotASungrowDeviceException()
    finally:
        await conn.close()



from pysungrow.definitions.device import SungrowDevice, SungrowDeviceType

# based on "Communication Protocol of PV Grid-Connected String Inverters" v1.1.2 dated 2017-5-13

sg60ktl = SungrowDevice(0x10F, "SG60KTL", SungrowDeviceType.STRING)
sg60ku = SungrowDevice(0x136, "SG60KU", SungrowDeviceType.STRING)
sg33ktl_m = SungrowDevice(0x134, "SG33KTL-M", SungrowDeviceType.STRING)
sg36ktl_m = SungrowDevice(0x74, "SG36KTL-M", SungrowDeviceType.STRING)
sg40ktl_m = SungrowDevice(0x135, "SG40KTL-M", SungrowDeviceType.STRING)
sg50ktl_m = SungrowDevice(0x11B, "SG50KTL-M", SungrowDeviceType.STRING)
sg60ktl_m = SungrowDevice(0x131, "SG60KTL-M", SungrowDeviceType.STRING)
sg60ku_m = SungrowDevice(0x132, "SG60KU-M", SungrowDeviceType.STRING)
sg49k5j = SungrowDevice(0x137, "SG49K5J", SungrowDeviceType.STRING)
sg8ktl_m = SungrowDevice(0x13F, "SG8KTL-M", SungrowDeviceType.STRING)
sg10ktl_m = SungrowDevice(0x13E, "SG10KTL-M", SungrowDeviceType.STRING)
sg12ktl_m = SungrowDevice(0x13C, "SG12KTL-M", SungrowDeviceType.STRING)
sg80ktl = SungrowDevice(0x138, "SG80KTL", SungrowDeviceType.STRING)
sg80ktl_m = SungrowDevice(0x139, "SG80KTL-M", SungrowDeviceType.STRING)
sg80hv = SungrowDevice(0x13A, "SG80HV", SungrowDeviceType.STRING)
sg125hv = SungrowDevice(0x13B, "SG125HV", SungrowDeviceType.STRING)

all_string = {
    sg10ktl_m,
    sg12ktl_m,
    sg125hv,
    sg33ktl_m,
    sg36ktl_m,
    sg40ktl_m,
    sg49k5j,
    sg50ktl_m,
    sg60ktl,
    sg60ktl_m,
    sg60ku,
    sg80hv,
    sg80ktl,
    sg80ktl_m,
    sg8ktl_m,
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysungrow',
 'pysungrow.definitions',
 'pysungrow.definitions.devices',
 'pysungrow.definitions.variables',
 'pysungrow.lib']

package_data = \
{'': ['*']}

install_requires = \
['pymodbus>=3.0.2,<4.0.0']

setup_kwargs = {
    'name': 'pysungrow',
    'version': '1.0.0',
    'description': 'Read and manipulate Sungrow inverters',
    'long_description': '# pysungrow - Python interface to Sungrow inverters\n\n![PyPI](https://img.shields.io/pypi/v/pysungrow)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pysungrow)\n![PyPI - Status](https://img.shields.io/pypi/status/pysungrow)\n![GitHub](https://img.shields.io/github/license/02jandal/pysungrow)\n[![CI](https://github.com/02JanDal/pysungrow/actions/workflows/ci.yaml/badge.svg?branch=main)](https://github.com/02JanDal/pysungrow/actions/workflows/ci.yaml)\n\nThis Python package provides abstractions over the Modbus protocol used by inverters of the brand Sungrow.\n\n## Features\n\n- Both getting and setting data\n- Fully async\n- Fully typed\n- Supports most forms of Modbus (TCP, UDP, TLS and Serial)\n- (Theoretically) supports most Sungrow inverters, both string and hybrid\n- High test coverage\n\n## Supported inverters\n\n**Tested:** SH10RT\n\n**In theory:** SG60KTL, SG60KU, SG33KTL-M, SG36KTL-M, SG40KTL-M, SG50KTL-M, SG60KTL-M, SG60KU-M, SG49K5J, SG8KTL-M, SG10KTL-M, SG12KTL-M, SG80KTL, SG80KTL-M, SG80HV, SG125HV, SH5K-20, SH3K6, SH4K6, SH5K-V13, SH5K-30, SH3K6-30, SH4K6-30, SH5.0RS, SH3.6RS, SH4.6RS, SH6.0RS, SH8.0RT, SH6.0RT, SH5.0RT\n\nDo you have an inverter that\'s not been tested yet? Please follow the instructions under _Getting started_ including running the `get` command, and report the result in a [new issue](https://github.com/02JanDal/pysungrow/issues/new).\n\n## Getting started\n\nInstall using `pip`:\n\n```bash\npip install pysungrow\n```\n\nSee below for usaging from Python. Also comes with a simple command line interface:\n\n```\npysungrow [-p PORT] [-s SLAVE] [HOST] identify\npysungrow [-p PORT] [-s SLAVE] [HOST] get [-k KEY]\npysungrow [-p PORT] [-s SLAVE] [HOST] set [KEY] [VALUE]\n```\n\nIt is recommended to start using these commands to verify that you can connect to your inverter successfully.\n\n## Usage\n\n### Getting data from the inverter\n\n```python\nfrom pysungrow import identify, SungrowClient\nfrom pymodbus.client import AsyncModbusTcpClient\n\nasync def example_get():\n    modbus_client = AsyncModbusTcpClient("192.168.1.228")\n\n    # first we need to identify the model of inverter...\n    serial_number, device, output_type = await identify(modbus_client)\n\n    # ...then we can create a client...\n    client = SungrowClient(modbus_client, device, output_type)\n\n    # ...using which we can get data\n    return await client.get("total_dc_power")\n```\n\nNote that the first call to `client.get` will fetch all variables defined for your model of inverter. You can limit this by first manually triggering a fetch using `await client.refresh(["total_dc_power"])`.\n\n### Controlling the inverter\n\n```python\nfrom pysungrow import identify, SungrowClient\nfrom pysungrow.definitions.variables.hybrid import ChargeDischargeCommand\nfrom pymodbus.client import AsyncModbusTcpClient\n\nasync def example_set():\n    modbus_client = AsyncModbusTcpClient("192.168.1.228")\n\n    # first we need to identify the model of inverter...\n    serial_number, device, output_type = await identify(modbus_client)\n\n    # ...then we can create a client...\n    client = SungrowClient(modbus_client, device, output_type)\n\n    # ...using which we can control the inverter\n    await client.set("charge_discharge_command", ChargeDischargeCommand.CHARGE)\n```\n\n## Contributing\n\nContributions are always welcome!\n\nFor code contributions please make sure that all automated checks pass. The easiest way to do this is using these commands:\n\n```bash\npre-commit run --all-files\npytest\n```\n## Acknowledgements\n\nThere are a few other similar projects available (however neither of them fit my needs):\n\n - [SungrowInverter](https://github.com/mvandersteen/SungrowInverter) by @mvandersteen\n - [HomeAssistant Modbus mappings](https://github.com/mkaiser/Sungrow-SHx-Inverter-Modbus-Home-Assistant) by @mkaiser\n - [sungrow-websocket](https://github.com/wallento/sungrow-websocket) by @wallento\n \n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Jan Dalheimer',
    'author_email': 'jan@dalheimer.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/02JanDal/pysungrow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Functions for decoding data from registers."""

import math
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union, cast

from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadDecoder

from pysungrow.definitions.variable import RawType, VariableDefinition

T = TypeVar("T")


def decode_registers_for_variable(
    decoder: BinaryPayloadDecoder, variable: VariableDefinition[T]
) -> Optional[T]:
    """
    Decode the information for the given variable.

    First decodes the raw data (which controls the number of registers used) before doing any
    additional transforms to Python data types.

    :param decoder: Decoder that contains the data, positioned at the start of the given register
    :param variable: The register for which to decode information
    :return: The decoded value
    """
    raw: Union[int, str, Tuple[int, ...]]
    if variable.raw_type == RawType.U16:
        raw = decoder.decode_16bit_uint()
    elif variable.raw_type == RawType.U32:
        raw = decoder.decode_32bit_uint()
    elif variable.raw_type == RawType.S16:
        raw = decoder.decode_16bit_int()
    elif variable.raw_type == RawType.S32:
        raw = decoder.decode_32bit_int()
    elif variable.raw_type == RawType.STRING:
        assert variable.length is not None
        raw = decoder.decode_string(variable.length * 2).decode("utf-8")
        if isinstance(raw, str) and "\x00" in raw:
            raw = raw[0 : raw.find("\x00")]
    elif variable.raw_type == RawType.RAW:
        raw = tuple(
            cast(int, decoder.decode_16bit_uint()) for _ in range(variable.length or 1)
        )
    else:  # pragma: no cover
        raise KeyError("Unknown raw type")

    if (
        isinstance(variable.transform, float)
        and variable.type == float
        and isinstance(raw, int)
    ):
        value = float(raw) * variable.transform

        # calculate number of decimals after the point based on scale (0.1 -> 1, 0.01 -> 2)
        after_point = -math.log10(variable.transform)
        # if the number of decimals is a whole number we can use it for rounding
        if after_point == round(after_point):
            value = round(value, int(after_point))

        return value  # type: ignore
    elif isinstance(variable.transform, int):
        return int(raw) * variable.transform
    elif isinstance(variable.transform, tuple):
        return variable.transform[0](raw)
    elif callable(variable.transform):
        return variable.transform(raw)  # type: ignore
    else:
        # also handles the case where register.transform is set to an enum
        return variable.type(raw)  # type: ignore


def decode_variables(
    data: List[int], variables: List[VariableDefinition]
) -> Dict[str, Any]:
    """
    Decode information for all variables fetched in a single read.

    :param data: The raw data as returned by pymodbus; a sequence of words (as in the 2 byte data type)
    :param variables: The variables for which the data was fetched
    :return: The decoded data for all passed registers
    """
    decoder = BinaryPayloadDecoder.fromRegisters(
        data, wordorder=Endian.Little, byteorder=Endian.Big
    )
    current = variables[0].address
    result: Dict[str, Any] = {}
    for variable in variables:
        skip = variable.address - current
        if skip > 0:
            decoder.skip_bytes(skip * 2)
        result[variable.key] = decode_registers_for_variable(decoder, variable)
        current = variable.end_address + 1
    return result



"""Functions for writing to Modbus clients."""

from typing import TypeVar

from pymodbus.client import ModbusBaseClient

from pysungrow.definitions.variable import VariableDefinition, VariableType
from pysungrow.lib.encode import encode_registers_for_variable

T = TypeVar("T")


async def write_variable(
    client: ModbusBaseClient, data: T, variable: VariableDefinition[T], slave: int
):
    """
    Write some data to a client.

    :param client: Client to which registers are written
    :param data: Data to write
    :param variable: Variable to write
    :param slave: Unit/slave to which registers are written
    """
    assert variable.variable_type == VariableType.READWRITE
    registers = encode_registers_for_variable(data, variable)
    # note: communication address = protocol address - 1
    await client.write_registers(variable.address - 1, registers, slave=slave)



"""Classes for defining a variable provided by a Sungrow inverter."""

from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Callable,
    Generic,
    Literal,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from pysungrow.definitions.device import SungrowDevice

if TYPE_CHECKING:
    from pysungrow.definitions.variables.device import OutputType


class RawType(str, Enum):
    """
    The "raw" type of data in a register (or sequence of registers).

    Can be further refined using the `transform` on the register (sequence).
    """

    U16 = "U16"
    U32 = "U32"
    S16 = "S16"
    S32 = "S32"
    STRING = "STRING"  # UTF-8
    RAW = "RAW"


class VariableType(str, Enum):
    """
    The type of variable.

    Can be:
    * Read-only (input register)
    * Read-only static (input register, doesn't change)
    * Read-write (holding register)
    """

    READ_STATIC = "READ_STATIC"
    READ = "READ"
    READWRITE = "READWRITE"


T = TypeVar("T")


@dataclass
class VariableDefinition(Generic[T]):
    """
    Stores information about a variable provided by a Sungrow inverter.

    Variables are stored in one or a consecutive sequence of registers (input or holding).

    This class contains the following information:
        * Metadata (`key`, `description`, `unit`)
        * Register address (`variable_type`, `address`)
        * Decoding/encoding instructions (`raw_type`, `type`, `transform`, `length`)
    """

    variable_type: VariableType
    key: str
    address: int
    raw_type: RawType
    type: Type[T]
    devices: Set[SungrowDevice]
    transform: Union[
        Type[T],
        float,
        Callable[[Union[int, str, Tuple[int, ...]]], Optional[T]],
        Tuple[
            Callable[[Union[int, str, Tuple[int, ...]]], Optional[T]],
            Callable[[T], Union[int, str, Tuple[int, ...]]],
        ],
        None,
    ] = None
    description: Optional[str] = None
    unit: Optional[
        Literal[
            "V",
            "A",
            "W",
            "kW",
            "Wh",
            "kWh",
            "°C",
            "Hz",
            "var",
            "K var",
            "kΩ",
            "%",
            "h",
            "s",
            "min",
            "kg",
            "Ah",
        ]
    ] = None
    length: Optional[int] = None
    limits: Optional[Union[Tuple[int, int], Tuple[float, float]]] = None
    if_output_type: Optional[Set["OutputType"]] = None

    @property
    def end_address(self):
        """Last register address which this variable covers."""
        return (
            self.address
            + {
                RawType.U16: 1,
                RawType.U32: 2,
                RawType.S16: 1,
                RawType.S32: 2,
                RawType.STRING: self.length,
                RawType.RAW: self.length,
            }[self.raw_type]
            - 1
        )
